// @module : Dummy_Cores_Mesh
// @author : Michel Kinsy
 /* 
 *  Using Heracles Multicore System under the copytright below:
 *  Copyright (c) 2011 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 *
 */

module Dummy_Cores_Mesh #(parameter ROW = 3, COLUMN = 3,  // size of the mesh 
						 VC_PER_PORTS = 2,  VC_DEPTH = 8, // size of the virtual channels specifications
						 RT_ALG = 0, EXTRA  = 2,          // routing algorithm 
						 DATA_WIDTH = 32,                 // out data width
						 TRAFFIC = 1, RATE = 5,           // traffic type and injection rate 
						 FLITS_PER_PACKET  = 8,           // flits per packet
						 STATS_CYCLES = 1)( // STATS_CYCLES is used to set number of cycles to collect 
						                // performance data
  clock,
  global_reset,
  
  start, stop,  
  operation, ON, core_ID, dest_ID,
  reset, PROG,
 
  route_table_address, route_table_data, 
  
  origin, data_out, valid_out  
  ); 
  
    //define the log2 function
	function integer log2;
	  input integer num;
	  integer i, result;
	  begin
		result = 1; 
		for (i = 0; 2 ** i < num; i = i + 1)
		  result = i + 1;
		log2 = result;
	  end
	endfunction
	
  // Change these local parameters only if you are familar with the system composition
  localparam IN_PORTS = 1, OUT_PORTS = 1, TYPE_BITS = 2, SWITCH_TO_SWITCH = 1 , OPER_BITS = 4; 

  //Number of bits to represent ID. 
  localparam ROW_BITS = log2(ROW); 
  localparam COLUMN_BITS = log2(COLUMN);
  localparam ID_BITS = ROW_BITS + COLUMN_BITS; 
  
  //Number of bits to represent vcs. 
  localparam VC_BITS = log2(VC_PER_PORTS); 
  
  //Number of bits to represent vc depth. 
  localparam VC_DEPTH_BITS = log2(VC_DEPTH);

  localparam  FLOW_BITS = (2*ID_BITS) + EXTRA;  
  localparam  FLIT_WIDTH = FLOW_BITS + TYPE_BITS + VC_BITS + DATA_WIDTH;

  localparam  PORTS = OUT_PORTS + 4*(SWITCH_TO_SWITCH);
  localparam  PORTS_BITS = log2(PORTS); 
  localparam  RT_WIDTH = PORTS_BITS + VC_BITS + 1; 
  
  localparam SWITCHES =  ROW * COLUMN;
  localparam CORES = SWITCHES;
// -------------------------------------------------------------------------------------


  input clock;
  input global_reset;
  input start, stop;
  
  input [OPER_BITS - 1: 0] operation;
  input ON; 
  input [ID_BITS - 1: 0] core_ID;
  input [ID_BITS - 1: 0] dest_ID;
  input reset; 
  input PROG;
 
  input [(FLOW_BITS - 1): 0]  route_table_address;
  input [RT_WIDTH - 1: 0]     route_table_data; 
  
  output [ID_BITS - 1: 0] origin;
  output [31: 0] data_out; 
  output valid_out;
 
  reg c_start  [0: CORES-1]; 
  reg c_stop   [0: CORES-1];
  reg [ID_BITS - 1: 0] c_dest_ID [0: CORES-1];
  
  reg  [1 :0]   from_peripheral       [0: CORES-1];
  reg  [31 :0]  from_peripheral_data  [0: CORES-1]; 
  reg           from_peripheral_valid [0: CORES-1]; 
  wire [1 :0]   to_peripheral      [0: CORES-1];
  wire [31 :0]  to_peripheral_data [0: CORES-1]; 
  wire          to_peripheral_valid[0: CORES-1];
    
  reg c_ON    [0: CORES-1];     
  reg c_reset [0: CORES-1]; 
  reg c_PROG  [0: CORES-1];
  reg c_WEn   [0: CORES-1];
  reg c_REn   [0: CORES-1];
  wire [ID_BITS - 1: 0] router_ID [0: CORES-1];
  reg [(FLOW_BITS - 1): 0] rt_flowID [0: SWITCHES-1]; 
  reg [RT_WIDTH - 1: 0] rt_entry [0: SWITCHES-1];
  
  
  wire [(FLIT_WIDTH * (SWITCH_TO_SWITCH *4)) - 1: 0] s_flits_in [0: SWITCHES-1];
  wire [(SWITCH_TO_SWITCH *4) - 1: 0]s_valid_in [0: SWITCHES-1]; 
  
  wire [(FLIT_WIDTH * (SWITCH_TO_SWITCH *4)) - 1: 0] s_flits_out [0: SWITCHES-1]; 
  wire [(SWITCH_TO_SWITCH *4) - 1: 0] s_valid_out [0: SWITCHES-1]; 
  
  wire [(((1 << VC_BITS) * (SWITCH_TO_SWITCH *4)) - 1): 0] s_empty_in  [0: SWITCHES-1]; 
  wire [(((1 << VC_BITS) * (SWITCH_TO_SWITCH *4)) - 1): 0] s_full_in  [0: SWITCHES-1]; 

  wire [(((1 << VC_BITS) * (SWITCH_TO_SWITCH *4)) - 1): 0] s_empty_out [0: SWITCHES-1];
  wire [(((1 << VC_BITS) * (SWITCH_TO_SWITCH *4)) - 1): 0] s_full_out [0: SWITCHES-1];  
  
  wire [(FLIT_WIDTH * SWITCH_TO_SWITCH) - 1: 0] single_s_flits_in [0: (SWITCHES*4)-1]; 
  wire [SWITCH_TO_SWITCH - 1: 0] single_s_valid_in [0: (SWITCHES*4)-1];  
  wire [(FLIT_WIDTH * SWITCH_TO_SWITCH) - 1: 0] single_s_flits_out [0: (SWITCHES*4)-1]; 
  wire [SWITCH_TO_SWITCH - 1: 0] single_s_valid_out [0: (SWITCHES*4)-1];  
  
  wire [(((1 << VC_BITS) * SWITCH_TO_SWITCH) - 1): 0] single_s_empty_in  [0: (SWITCHES*4)-1]; 
  wire [(((1 << VC_BITS) * SWITCH_TO_SWITCH) - 1): 0] single_s_full_in  [0: (SWITCHES*4)-1];
  wire [(((1 << VC_BITS) * SWITCH_TO_SWITCH) - 1): 0] single_s_empty_out  [0: (SWITCHES*4)-1]; 
  wire [(((1 << VC_BITS) * SWITCH_TO_SWITCH) - 1): 0] single_s_full_out  [0: (SWITCHES*4)-1];
  
  genvar i;
  generate
	for (i = 0; i < SWITCHES; i= i + 1) begin :  S_IN_OUT
 		 assign  s_flits_in[i][((FLIT_WIDTH * SWITCH_TO_SWITCH)-1) -: (FLIT_WIDTH * SWITCH_TO_SWITCH)] = single_s_flits_in[(i*4)] ;
 		 assign  s_valid_in[i][(SWITCH_TO_SWITCH-1) -: SWITCH_TO_SWITCH] = single_s_valid_in[(i*4)];
 		 assign  s_flits_in[i][((2*(FLIT_WIDTH * SWITCH_TO_SWITCH))-1) -: (FLIT_WIDTH * SWITCH_TO_SWITCH)] = single_s_flits_in[((i*4)+1)];
 		 assign  s_valid_in[i][((2*SWITCH_TO_SWITCH)-1) -: SWITCH_TO_SWITCH] = single_s_valid_in[((i*4)+1)]; 
 		 assign  s_flits_in[i][((3*(FLIT_WIDTH * SWITCH_TO_SWITCH))-1) -: (FLIT_WIDTH * SWITCH_TO_SWITCH)] = single_s_flits_in[((i*4)+2)] ;
 		 assign  s_valid_in[i][((3*SWITCH_TO_SWITCH)-1) -: SWITCH_TO_SWITCH] = single_s_valid_in[((i*4)+2)] ;
 		 assign  s_flits_in[i][((4*(FLIT_WIDTH * SWITCH_TO_SWITCH))-1) -: (FLIT_WIDTH * SWITCH_TO_SWITCH)] = single_s_flits_in[((i*4)+3)];
 		 assign  s_valid_in[i][((4*SWITCH_TO_SWITCH)-1) -: SWITCH_TO_SWITCH] = single_s_valid_in[((i*4)+3)] ;
 		 
 		 assign  single_s_flits_out[(i*4)] =  s_flits_out[i][((FLIT_WIDTH * SWITCH_TO_SWITCH)-1) -: (FLIT_WIDTH * SWITCH_TO_SWITCH)];
 		 assign  single_s_valid_out[(i*4)] =  s_valid_out[i][(SWITCH_TO_SWITCH-1) -: SWITCH_TO_SWITCH];
 		 assign  single_s_flits_out[((i*4)+1)] =  s_flits_out[i][((2*(FLIT_WIDTH * SWITCH_TO_SWITCH))-1) -: (FLIT_WIDTH * SWITCH_TO_SWITCH)];
 		 assign  single_s_valid_out[((i*4)+1)] =  s_valid_out[i][((2*SWITCH_TO_SWITCH)-1) -: SWITCH_TO_SWITCH];
 		 assign  single_s_flits_out[((i*4)+2)] =  s_flits_out[i][((3*(FLIT_WIDTH * SWITCH_TO_SWITCH))-1) -: (FLIT_WIDTH * SWITCH_TO_SWITCH)];
 		 assign  single_s_valid_out[((i*4)+2)] =  s_valid_out[i][((3*SWITCH_TO_SWITCH)-1) -: SWITCH_TO_SWITCH];
 		 assign  single_s_flits_out[((i*4)+3)] =  s_flits_out[i][((4*(FLIT_WIDTH * SWITCH_TO_SWITCH))-1) -: (FLIT_WIDTH * SWITCH_TO_SWITCH)];
 		 assign  single_s_valid_out[((i*4)+3)] =  s_valid_out[i][((4*SWITCH_TO_SWITCH)-1) -: SWITCH_TO_SWITCH];	
 		 
 
  		 assign  s_empty_in[i][(((1 << VC_BITS) * SWITCH_TO_SWITCH)-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)] = single_s_empty_in[(i*4)] ;
 		 assign  s_empty_in[i][((2*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)] = single_s_empty_in[((i*4)+1)];
 		 assign  s_empty_in[i][((3*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)] = single_s_empty_in[((i*4)+2)] ;
 		 assign  s_empty_in[i][((4*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)] = single_s_empty_in[((i*4)+3)];

  		 assign  s_full_in[i][(((1 << VC_BITS) * SWITCH_TO_SWITCH)-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)] = single_s_full_in[(i*4)] ;
 		 assign  s_full_in[i][((2*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)] = single_s_full_in[((i*4)+1)];
 		 assign  s_full_in[i][((3*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)] = single_s_full_in[((i*4)+2)] ;
 		 assign  s_full_in[i][((4*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)] = single_s_full_in[((i*4)+3)];
 		  		 
 		 assign  single_s_empty_out[(i*4)]      =  s_empty_out[i][(((1 << VC_BITS) * SWITCH_TO_SWITCH)-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)];
 		 assign  single_s_empty_out[((i*4)+1)]  =  s_empty_out[i][((2*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)];
 		 assign  single_s_empty_out[((i*4)+2)]  =  s_empty_out[i][((3*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)];
 		 assign  single_s_empty_out[((i*4)+3)]  =  s_empty_out[i][((4*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)];  
 		 
 		 assign  single_s_full_out[(i*4)]      =  s_full_out[i][(((1 << VC_BITS) * SWITCH_TO_SWITCH)-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)];
 		 assign  single_s_full_out[((i*4)+1)]  =  s_full_out[i][((2*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)];
 		 assign  single_s_full_out[((i*4)+2)]  =  s_full_out[i][((3*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)];
 		 assign  single_s_full_out[((i*4)+3)]  =  s_full_out[i][((4*((1 << VC_BITS) * SWITCH_TO_SWITCH))-1) -: ((1 << VC_BITS) * SWITCH_TO_SWITCH)];  
 		  	    
	end
  endgenerate  

 generate
  	assign  single_s_flits_in[0] = 0;
 	assign  single_s_valid_in[0] = 0;
  	assign  single_s_empty_in[0] = 0;
 	assign  single_s_full_in[0]  = 0;
 	for (i = 1; i < SWITCHES ; i= i + 1) begin :  SS_IN
 		 assign  single_s_flits_in[(i*4)] =  (i % COLUMN)? single_s_flits_out[(((i-1)*4)+ 2)] : 0;
 		 assign  single_s_valid_in[(i*4)] =  (i % COLUMN)? single_s_valid_out[(((i-1)*4)+ 2)] : 0;
  		 assign  single_s_empty_in[(i*4)] =  (i % COLUMN)? single_s_empty_out[(((i-1)*4)+ 2)] : 0;
 		 assign  single_s_full_in[(i*4)]  =  (i % COLUMN)?  single_s_full_out[(((i-1)*4)+ 2)] : 0;
    end 
 
 	for (i = 0; i < SWITCHES - COLUMN ; i= i + 1) begin :  SS_IN_2
 		 assign  single_s_flits_in[((i*4)+1)] =  single_s_flits_out[(((i+COLUMN)*4)+ 3)];
 		 assign  single_s_valid_in[((i*4)+1)] =  single_s_valid_out[(((i+COLUMN)*4)+ 3)];  
 		 assign  single_s_empty_in[((i*4)+1)] =  single_s_empty_out[(((i+COLUMN)*4)+ 3)];
 		 assign  single_s_full_in[((i*4)+1)]  =  single_s_full_out[(((i+COLUMN)*4)+ 3)]; 	    
	end	
	for (i = SWITCHES - COLUMN; i < SWITCHES ; i= i + 1) begin :  SS_IN_3 
 		 assign  single_s_flits_in[((i*4)+1)] =  0;
 		 assign  single_s_valid_in[((i*4)+1)] =  0;  
  		 assign  single_s_empty_in[((i*4)+1)] =  0;
 		 assign  single_s_full_in[((i*4)+1)]  =  0;  	    
	end	
	
	for (i = 0; i < SWITCHES - 1 ; i= i + 1) begin :  SS_IN_4
 		 assign  single_s_flits_in[((i*4)+2)] =  ((i+1) % COLUMN)? single_s_flits_out[((i+1)*4)] : 0;
 		 assign  single_s_valid_in[((i*4)+2)] =  ((i+1) % COLUMN)? single_s_valid_out[((i+1)*4)] : 0;	
  		 assign  single_s_empty_in[((i*4)+2)] =  ((i+1) % COLUMN)? single_s_empty_out[((i+1)*4)] : 0;
 		 assign  single_s_full_in[((i*4)+2)] =  ((i+1) % COLUMN)? single_s_full_out[((i+1)*4)] : 0;   	    
	end

	for (i = 0; i < COLUMN ; i= i + 1) begin :  SS_IN_5 
 		 assign  single_s_flits_in[((i*4)+3)] = 0;
 		 assign  single_s_valid_in[((i*4)+3)] = 0; 	 
  		 assign  single_s_empty_in[((i*4)+3)] = 0;
 		 assign  single_s_full_in[((i*4)+3)] = 0; 	   
	end	
	for (i = COLUMN; i < SWITCHES ; i= i + 1) begin :  SS_IN_6 
 		 assign  single_s_flits_in[((i*4)+3)] = single_s_flits_out[(((i-COLUMN)*4)+ 1)];
 		 assign  single_s_valid_in[((i*4)+3)] = single_s_valid_out[(((i-COLUMN)*4)+ 1)];  	
  		 assign  single_s_empty_in[((i*4)+3)] = single_s_empty_out[(((i-COLUMN)*4)+ 1)];
 		 assign  single_s_full_in[((i*4)+3)]  = single_s_full_out[(((i-COLUMN)*4)+ 1)];     
	end		
  endgenerate   
 
  generate
	for (i = 0; i < SWITCHES ; i= i + 1) begin : MESH_NODE
	     assign router_ID[i] = i;
		 Dummy_node #(ID_BITS, DATA_WIDTH, IN_PORTS, OUT_PORTS, VC_BITS, 
					 VC_DEPTH_BITS, EXTRA, TYPE_BITS, SWITCH_TO_SWITCH,
					 RT_ALG, ROW, COLUMN, TRAFFIC, RATE, FLITS_PER_PACKET, STATS_CYCLES) NODES (
                               clock, c_reset[i], 
  							   c_start[i], c_stop[i],
							   c_dest_ID[i], 

                               from_peripheral[i],
                               from_peripheral_data[i], 
                               from_peripheral_valid[i], 
                               to_peripheral[i],
                               to_peripheral_data[i], 
                               to_peripheral_valid[i],
                                                 
  							   c_ON [i], c_PROG[i], 
  							   c_WEn[i], c_REn[i], 
  							   
  							   router_ID[i], 
                               rt_flowID[i], rt_entry [i],  
  							   
                               s_flits_in[i], s_valid_in[i],  
                               s_flits_out[i], s_valid_out[i], 
                               s_empty_in[i], s_full_in[i],  
                               s_empty_out[i], s_full_out[i]
		);		       
	end
  endgenerate  

  reg  out_rdEn [0:CORES-1];
  wire out_read_valid [0:CORES-1];
  wire [31: 0] out_read_data [0:CORES-1];
  wire out_full [0:CORES-1];
  wire out_empty [0:CORES-1];
 
  generate
	for (i = 0; i < CORES ; i= i + 1) begin : OUT_DATA_BUFS
		fifo  #(32, VC_DEPTH_BITS, 0) Bufs (
			clock,
			c_reset[i],
			to_peripheral_data[i], 
			to_peripheral_valid[i], 
			out_rdEn[i],
			1'b0, 
 
			out_read_data [i], 
			out_read_valid[i],			
			out_full[i], 
			out_empty[i] 
		);
	end
  endgenerate  
 
 reg [31: 0] cycles;
 integer index; 
 
 reg [ID_BITS - 1: 0] cur_read; 
 reg [ID_BITS - 1: 0] prev_read;

 always @ (posedge clock) begin 
 	 	if(global_reset == 1)begin
			cycles   <= 1;
			cur_read <= 0;
			prev_read <= 0;
			for (index = 0; index < CORES; index = index+1) out_rdEn[index] <= 0; 		
		end
		else begin
			cycles   <= cycles + 1;
			cur_read <= (cur_read == (CORES - 1))? 0 : cur_read + 1;
			prev_read <= cur_read;
			if (prev_read != cur_read) begin 
			   out_rdEn[cur_read]  <= (out_empty[cur_read] || ~c_ON[cur_read])? 0 : 1;
			   out_rdEn[prev_read] <= 0;
			end
			else
				 out_rdEn[cur_read] <= (out_empty[cur_read] || ~c_ON[cur_read])? 0 : 1; 	
		end 
  end   	
 
    always @ (posedge clock) begin
		  if(global_reset == 1)begin
			   for (index = 0; index < CORES; index = index+1) begin
			       c_ON[index] <= 0;
			   end 		
		  end
		  else begin
            if(operation[0] == 1) begin
                 c_ON  [core_ID]   <=  ON;
            	 c_WEn [core_ID]   <= (ON == 1)? 1 : 0;
            	 c_REn [core_ID]   <= (ON == 1)? 1 : 0;
            end  
            if(operation[1] == 1) c_reset [core_ID] <= reset;
            if(operation[2] == 1) c_PROG  [core_ID] <= PROG;
            if(operation[3] == 1) begin
                c_start[core_ID]   <= start;
                c_stop[core_ID]    <= stop;
				c_dest_ID[core_ID] <= dest_ID; 
            end
      
            rt_flowID [core_ID] <= route_table_address;
            rt_entry [core_ID]  <= route_table_data;
		  end
    end
		  assign origin     = prev_read; 
		  assign valid_out  = c_ON[prev_read]? out_read_valid [prev_read]: 0; 
		  assign data_out   = c_ON[prev_read]? out_read_data[prev_read]:0; 
endmodule
