/** @module : Dir_cache_translator
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

 module cache_translator #(parameter CORE = 0, DATA_WIDTH = 32, OFFSET_BITS = 3, ADDRESS_BITS = 20,  MSG_BITS = 4)(
                     clock, reset,
                     
					 read, write, flush, invalid, 
					 share, address, in_line,
					 
				     invalidate, sharing, sharing_req, write_back, 
					 exclusive, modify, out_addr, out_line,
					 
				     in_msg, in_address, in_data_line, 
				     out_msg, out_address, out_data_line
 ); 

localparam OFFSET   = 1<<OFFSET_BITS; 
localparam BLOCK_BITS  = DATA_WIDTH * OFFSET; 

input clock, reset; 

input read, write, flush, invalid, share;
input [ADDRESS_BITS-1:0]  address; 
input [BLOCK_BITS-1:0]    in_line;
 				     
output  invalidate, sharing, sharing_req, write_back, exclusive, modify;
output [ADDRESS_BITS-1:0]  out_addr; 
output [BLOCK_BITS-1:0]    out_line;

input [MSG_BITS-1:0]      in_msg; 
input [ADDRESS_BITS-1:0]  in_address; 
input [BLOCK_BITS-1:0]    in_data_line;
 				     
output [MSG_BITS-1:0]     out_msg; 
output [ADDRESS_BITS-1:0] out_address; 
output [BLOCK_BITS-1:0]   out_data_line;

localparam NO_MSG = 0;
localparam M_REQ = 1, INV_REQ = 2, M_WB_REQ = 3, SHARE_REQ = 4, WB_INV_REQ = 5; //memory requests
localparam M_READY = 6, MEM_SENT = 7,  W_RESP = 8, EX_RESP = 9, SH_RESP = 10;   //memory responses 
localparam R_REQ = 1, W_REQ = 2, C_WB_REQ = 3;                                  //cache requests
localparam C_READY = 4, C_SENT = 5, C_RECV = 6, INV_RESP = 7, WB_RESP = 8, WB_INV_RESP = 9;
//SH_RESP = 10 //cache responses 

reg invalidate, sharing, sharing_req, write_back, exclusive, modify;
reg [ADDRESS_BITS-1:0]  out_addr; 
reg [BLOCK_BITS-1:0]    out_line;

reg [MSG_BITS-1:0]      out_msg; 
reg [ADDRESS_BITS-1:0]  out_address;
reg [BLOCK_BITS-1:0]    out_data_line; 

always @ (posedge clock) begin
	if (reset) begin
		invalidate   <= 0; 
		sharing      <= 0;  
		sharing_req  <= 0;
		write_back   <= 0;
		exclusive    <= 0;  
		modify       <= 0;
		out_addr     <= 0; 
		out_line     <= 0; 
	end 
	else begin
		case (in_msg)
			NO_MSG: begin
				invalidate   <= 0; 
				sharing      <= 0; 
				sharing_req  <= 0;				
				write_back   <= 0;
				exclusive    <= 0;  
				modify       <= 0;
				out_addr     <= 0;
				out_line     <= 0;
			end
			M_WB_REQ: begin
				invalidate   <= 0; 
				sharing      <= 0; 
				sharing_req  <= 0;	
				write_back   <= 1;
				exclusive    <= 0;  
				modify       <= 0;
				out_addr     <= in_address;
				out_line     <= 0;
			end
			INV_REQ: begin
				invalidate   <= 1; 
				sharing      <= 0;  
				sharing_req  <= 0;	
				write_back   <= 0;
				exclusive    <= 0;  
				modify       <= 0;
				out_addr     <= in_address;
				out_line     <= 0;
			end
			SHARE_REQ: begin
				invalidate   <= 0; 
				sharing      <= 0;  
				sharing_req  <= 1;	
				write_back   <= 0;
				exclusive    <= 0;  
				modify       <= 0;
				out_addr     <= in_address;
				out_line     <= 0;
			end
			WB_INV_REQ: begin
				invalidate   <= 1; 
				sharing      <= 0; 
				sharing_req  <= 0;					
				write_back   <= 1;
				exclusive    <= 0;  
				modify       <= 0;
				out_addr     <= in_address;
				out_line     <= 0;
			end
			W_RESP: begin
				invalidate   <= 0; 
				sharing      <= 0; 
				sharing_req  <= 0;					
				write_back   <= 0;
				exclusive    <= 0;  
				modify       <= 1;
				out_addr     <= in_address;
				out_line     <= in_data_line;
			end
			EX_RESP: begin
				invalidate   <= 0; 
				sharing      <= 0; 
				sharing_req  <= 0;					
				write_back   <= 0;
				exclusive    <= 1;  
				modify       <= 0;
				out_addr     <= in_address;
				out_line     <= in_data_line;
			end
			SH_RESP: begin
				invalidate   <= 0; 
				sharing      <= 1; 
				sharing_req  <= 0;					
				write_back   <= 0;
				exclusive    <= 0;  
				modify       <= 0;
				out_addr     <= in_address;
				out_line     <= in_data_line;
			end
			default: begin 
				invalidate   <= 0; 
				sharing      <= 0;
				sharing_req  <= 0;					
				write_back   <= 0;
				exclusive    <= 0;  
				modify       <= 0;
				out_addr     <= 0;
				out_line     <= 0;
			end 
		endcase
	end
end

always @ (posedge clock) begin
	if (reset) begin
		out_msg      <= NO_MSG; 
		out_address  <= 0;
	end 
	else begin
		case ({read, write, flush, invalid, share})
			5'b10000: begin 
				out_msg       <= R_REQ; 
				out_address   <= address;
				out_data_line <= 0; 
			end 
			5'b01000: begin 
				out_msg       <= W_REQ; 
				out_address   <= address;
				out_data_line <= 0; 
			end 
			5'b00100: begin 
				//out_msg       <= C_WB_REQ;
				out_msg       <= WB_RESP;				
				out_address   <= address;
				out_data_line <= in_line; 
			end 
			5'b00101: begin 
				out_msg       <= WB_RESP; 
				out_address   <= address;
				out_data_line <= in_line;
			end 
			5'b00110: begin 
				out_msg       <= WB_INV_RESP; 
				out_address   <= address;
				out_data_line <= in_line; 
			end 
			5'b00111: begin 
				out_msg       <= WB_INV_RESP; 
				out_address   <= address;
				out_data_line <= in_line; 
			end 
			5'b00010: begin 
				out_msg       <= INV_RESP; 
				out_address   <= address;
				out_data_line <= 0;
			end 
			5'b00001: begin 
				out_msg       <= SH_RESP; 
				out_address   <= address;
				out_data_line <= 0;
			end 
			default: begin 
				out_msg       <= NO_MSG; 
				out_address   <= address;
				out_data_line <= 0;
			end 
		endcase
	end
end

/*
always @ (posedge clock) begin 
	$display ("-------------------------- Cache Directory %d --------------------------------------", CORE);
	$display ("Read [%b]\t| Write [%b]\t| Flush [%b]\t| Invalid [%b]\t| Share [%b]\t| Address [%h]", read, write, flush, invalid, share, address);
	$display ("Invalidate [%b]| Sharing [%b]\t| Write_back [%b]| Exclusive [%b]\t| Modify[%b]\t| Address [%h]", invalidate, sharing, write_back, exclusive, modify, out_addr);
	$display ("IN_MSG [%d]\t| Address [%h]| Dataline [%h]", in_msg, in_address, in_data_line);
	$display ("OUT_MSG [%d]\t| Address [%h]| Dataline [%h]", out_msg, out_address, out_data_line);
	$display ("------------------------------------------------------------------------------------");
end 
//*/

endmodule