/** @module : main_memory
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
 
module main_memory #(parameter CORE = 0, DATA_WIDTH = 32, ADDRESS_BITS = 20, 
					 ID_BITS = 4, SHARERS = 4, 
					 REAL_ADDR_BITS = 16, MSG_BITS = 4, OFFSET_BITS = 3, 
					 INIT_FILE = "memory.mem") (
				     clock, reset,  
				     				     
				     cache2mem_msg, 
				     cache2mem_address, cache2mem_data,
				     mem2cache_msg, 
				     mem2cache_address, mem2cache_data
); 

input clock, reset; 				     
input [MSG_BITS-1:0]     cache2mem_msg; 
input [ADDRESS_BITS-1:0] cache2mem_address; 
input [DATA_WIDTH-1:0]   cache2mem_data;

output [MSG_BITS-1:0]     mem2cache_msg; 
output [ADDRESS_BITS-1:0] mem2cache_address; 
output [DATA_WIDTH-1:0]   mem2cache_data;

//define the log2 function
function integer log2;
  input integer num;
  integer i, result;
  begin
	for (i = 0; 2 ** i < num; i = i + 1)
	  result = i + 1;
	log2 = result;
  end
endfunction

localparam NUM_SHARERS_BITS = log2(SHARERS); 

wire    read, write; 
wire   [REAL_ADDR_BITS-1:0] r_address;
wire   [REAL_ADDR_BITS-1:0] w_address;
wire   [DATA_WIDTH-1:0]   data_in;
wire   [DATA_WIDTH-1:0]   data_out;

wire [MSG_BITS-1:0]        in_msg; 
wire [ID_BITS-1:0]         src_id;
wire [ADDRESS_BITS-1:0]    in_address; 

wire dir_rp_ready, dir_rq_ready; 				     
wire [MSG_BITS-1:0]       out_msg; 
wire [ID_BITS-1:0]        dest_id;
wire [ADDRESS_BITS-1:0]   out_address; 
	     
BRAM #(DATA_WIDTH, REAL_ADDR_BITS) RAM_Block(
		.clock(clock),
    	.readEnable(read),
    	.readAddress(r_address[(REAL_ADDR_BITS -1) -: REAL_ADDR_BITS]),
   	    .readData(data_out),
    	.writeEnable(write),
    	.writeAddress(w_address[(REAL_ADDR_BITS -1) -: REAL_ADDR_BITS]),
    	.writeData(data_in)
);

directory #(CORE, REAL_ADDR_BITS, MSG_BITS, ID_BITS, NUM_SHARERS_BITS) DIR (
		.clock(clock), 
		.reset(reset),
		.resp_ready(dir_rp_ready), 
		.req_ready(dir_rq_ready),

		.in_msg(in_msg), 
		.src_id(src_id),
		.in_address(in_address[(REAL_ADDR_BITS -1) -: REAL_ADDR_BITS]), 

		.out_msg(out_msg),
		.dest_id(dest_id),
		.out_address(out_address[(REAL_ADDR_BITS -1) -: REAL_ADDR_BITS])
 );
 
main_memory_translator #(CORE, DATA_WIDTH, ADDRESS_BITS, REAL_ADDR_BITS, ID_BITS, MSG_BITS, 
						OFFSET_BITS) translator (
		.clock(clock), 
		.reset(reset),

		//directory 
		.resp_ready(dir_rp_ready), 
		.req_ready(dir_rq_ready),
		.in_msg(out_msg), 
		.src_id(dest_id),
		.in_address(out_address[(REAL_ADDR_BITS -1) -: REAL_ADDR_BITS]), 

		.out_msg(in_msg),
		.dest_id(src_id),
		.out_address(in_address[(REAL_ADDR_BITS -1) -: REAL_ADDR_BITS]),

		//memory block 
    	.readEnable(read),
    	.readAddress(r_address[(REAL_ADDR_BITS -1) -: REAL_ADDR_BITS]),
   	    .readData(data_out),
    	.writeEnable(write),
    	.writeAddress(w_address[(REAL_ADDR_BITS -1) -: REAL_ADDR_BITS]),
    	.writeData(data_in),

		//main interface 
		.cache2mem_msg(cache2mem_msg),
		.cache2mem_address(cache2mem_address),
		.cache2mem_data(cache2mem_data),
		.mem2cache_msg(mem2cache_msg),
		.mem2cache_address(mem2cache_address),
		.mem2cache_data(mem2cache_data)
); 

endmodule
