/** @module : Memory splitter
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2013 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

 module memory_splitter #(parameter CORE = 0, DATA_WIDTH = 32, INDEX_BITS = 6, 
                     OFFSET_BITS = 3, ADDRESS_BITS = 20, CACHE_ID_BITS = 5, 
					 ID_BITS = 4, MSG_BITS = 4)(
                     clock, reset,

					 caches2mem_msg, 
				     caches2mem_address, caches2mem_data,
					 
				     mem2caches_msg, 
				     mem2caches_address, mem2caches_data, 
					 
					 local2cache_msg, 
				     local2cache_address, local2cache_data,

					 cache2local_msg, 
				     cache2local_address, cache2local_data,

					 network2mem_msg, 
				     network2mem_address, network2mem_data,
					 
					 mem2network_msg, 
				     mem2network_address, mem2network_data
 ); 

input clock, reset; 

input [MSG_BITS-1:0]       mem2caches_msg; 
input [ADDRESS_BITS-1:0]   mem2caches_address; 
input [DATA_WIDTH-1:0]     mem2caches_data;

output [MSG_BITS-1:0]      caches2mem_msg; 
output [ADDRESS_BITS-1:0]  caches2mem_address; 
output [DATA_WIDTH-1:0]    caches2mem_data;

input [MSG_BITS-1:0]       cache2local_msg; 
input [ADDRESS_BITS-1:0]   cache2local_address; 
input [DATA_WIDTH-1:0]     cache2local_data;

output [MSG_BITS-1:0]      local2cache_msg; 
output [ADDRESS_BITS-1:0]  local2cache_address; 
output [DATA_WIDTH-1:0]    local2cache_data;

input [MSG_BITS-1:0]       network2mem_msg; 
input [ADDRESS_BITS-1:0]   network2mem_address; 
input [DATA_WIDTH-1:0]     network2mem_data;

output [MSG_BITS-1:0]      mem2network_msg; 
output [ADDRESS_BITS-1:0]  mem2network_address; 
output [DATA_WIDTH-1:0]    mem2network_data;

localparam INFO_WIDTH = MSG_BITS + ADDRESS_BITS + DATA_WIDTH; 
localparam NO_MSG = 0;
localparam M_REQ = 1, INV_REQ = 2, M_WB_REQ = 3, SHARE_REQ = 4, WB_INV_REQ = 5; //memory requests
localparam M_READY = 6, MEM_SENT = 7,  W_RESP = 8, EX_RESP = 9, SH_RESP = 10;   //memory responses 
localparam R_REQ = 1, W_REQ = 2, C_WB_REQ = 3;                                  //cache requests
localparam C_READY = 4, C_SENT = 5, C_RECV = 6, INV_RESP = 7, WB_RESP = 8, WB_INV_RESP = 9;
//cache responses

localparam REAL_ADDR_BITS = ADDRESS_BITS - ID_BITS; 

wire [INFO_WIDTH-1:0] network_info = {network2mem_msg, network2mem_address, network2mem_data}; 
wire write_n_info = (network2mem_msg != NO_MSG); 

wire [INFO_WIDTH-1:0] cache_info = {cache2local_msg, cache2local_address, cache2local_data}; 
wire write_c_info = (cache2local_msg != NO_MSG);
 
wire valid;
wire [INFO_WIDTH - 1:0] read_data; 
wire [ADDRESS_BITS-1:0] read_addr = read_data[(INFO_WIDTH -MSG_BITS-1) -: ADDRESS_BITS]; 

wire [ID_BITS -1: 0]     core_id = CORE; 
wire [ID_BITS -1: 0]     mem_addr_pad = 0;
wire [ADDRESS_BITS-1:0]  t_mem2caches_address  = {core_id, mem2caches_address[REAL_ADDR_BITS-1 :0]}; 
wire [ADDRESS_BITS-1:0]  t_caches2mem_address  = {mem_addr_pad, read_addr[REAL_ADDR_BITS-1 :0]}; 

reg  [CACHE_ID_BITS -1: 0] t_dest; 
wire [CACHE_ID_BITS -1: 0] dest =((mem2caches_msg == INV_REQ) | (mem2caches_msg == WB_INV_REQ) | 
							   (mem2caches_msg == SHARE_REQ) | (mem2caches_msg == M_WB_REQ) | 
							   (mem2caches_msg == W_RESP) | (mem2caches_msg == EX_RESP) | 
							   (mem2caches_msg == SH_RESP) | (mem2caches_msg == M_READY))?
							    mem2caches_data : t_dest; 

wire local = (dest[(CACHE_ID_BITS -1) -: ID_BITS] == core_id); 

merger #(INFO_WIDTH, 2, 3, 0) MEMM (
  .clock(clock), 
  .reset(reset), 
  .write({write_n_info, write_c_info}), 
  .write_data({network_info, cache_info}), 

  .read(1'b1), 
  .read_data(read_data), 
  .out_fifo(), 
  .valid(valid), 
  
  .empty(), 
  .full()
); 

always @ (posedge clock) begin
	t_dest     <= reset? 0: dest; 
end 

//----------------------------------------------------
// Drive the outputs
//----------------------------------------------------
assign caches2mem_msg       = reset? 0 : valid? read_data[INFO_WIDTH-1 -: MSG_BITS]: NO_MSG; 
assign caches2mem_address   = reset? 0 : valid? t_caches2mem_address : 0;  
assign caches2mem_data      = reset? 0 : valid? read_data[DATA_WIDTH-1 : 0]: 0; 

assign local2cache_msg      = ((mem2caches_msg != NO_MSG) & (local))? mem2caches_msg : NO_MSG; 
assign local2cache_address  = ((mem2caches_msg != NO_MSG) & (local))? t_mem2caches_address : 0; 
assign local2cache_data     = ((mem2caches_msg != NO_MSG) & (local))? mem2caches_data : 0; 

assign mem2network_msg      = ((mem2caches_msg != NO_MSG) & (~local))? mem2caches_msg : NO_MSG; 
assign mem2network_address  = ((mem2caches_msg != NO_MSG) & (~local))? t_mem2caches_address : 0; 
assign mem2network_data     = ((mem2caches_msg != NO_MSG) & (~local))? mem2caches_data : 0;

/*
 always @ (posedge clock) begin 
	$display ("--------------------------------Memory Splitter %d ----------------------------------", CORE); 
	$display ("caches2mem_msg [%d]\t| caches2mem_address [%h]   | caches2mem_data [%h]", caches2mem_msg, caches2mem_address,caches2mem_data);
	$display ("mem2caches_msg [%d]\t| mem2caches_address [%h]   | mem2caches_data [%h]", mem2caches_msg, mem2caches_address,mem2caches_data);

	$display ("local2cache_msg [%d]\t| local2cache_address [%h]| local2cache_data [%h]", local2cache_msg, local2cache_address,local2cache_data);
	$display ("cache2local_msg [%d]\t| cache2local_address [%h]| cache2local_data [%h]", cache2local_msg, cache2local_address,cache2local_data); 

	$display ("network2mem_msg [%d]\t| network2mem_address [%h]  | network2mem_data [%h]", network2mem_msg, network2mem_address,network2mem_data);
	$display ("mem2network_msg [%d]\t| mem2network_address [%h]  | mem2network_data [%h]", mem2network_msg, mem2network_address,mem2network_data);
	$display ("-----------------------------------------------------------------------------------");			 
 end
//*/

endmodule 