
/** @module : merger
 *  @author : Michel Kinsy
 
 *  Copyright (c) 2012   Michel Kinsy (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
//`include "fifo.v" 
 
module merger #(parameter DATA_WIDTH = 32, NUM_FIFO = 2, Q_DEPTH_BITS = 3, Q_IN_BUFFERS = 2) (
    clock, reset,
    
    write,
	write_data,

    read,	
	read_data, 
	out_fifo, 
	valid, 
	
	empty, full 
); 

//define the log2 function
function integer log2;
  input integer num;
  integer i, result;
  begin
	for (i = 0; 2 ** i < num; i = i + 1)
	  result = i + 1;
	log2 = result;
  end
endfunction

localparam FIFO_BITS = log2(NUM_FIFO); 

input clock;
input reset;
input [NUM_FIFO - 1:0] write;
input [(DATA_WIDTH * NUM_FIFO) - 1:0]  write_data;

input                     read;
output [DATA_WIDTH - 1:0] read_data; 
output [FIFO_BITS - 1:0]  out_fifo;  
output valid; 
	
output [NUM_FIFO -1:0] empty; 
output [NUM_FIFO -1:0] full; 
	
wire [DATA_WIDTH - 1:0] b_write_data [0:NUM_FIFO-1];
wire [DATA_WIDTH - 1:0] b_read_data [0:NUM_FIFO-1];
wire [NUM_FIFO - 1:0]   b_valid; 
reg  [NUM_FIFO - 1:0]   b_read;
  
function  [FIFO_BITS-1:0] get_fifo;
	input [NUM_FIFO-1:0] empty;
	integer u, index; 
	integer found;
	begin
		found = 0; index = 0; 
		for(u= 0; u < NUM_FIFO; u = u+1) begin
			if ((empty[u] == 0) & (found == 0))begin 
				found = 1; 
				index = u; 
			end 			
		end 
		get_fifo = index;
	end
endfunction  
  
genvar i;
generate
for (i = 0; i < NUM_FIFO; i=i+1) begin : MFFCL
	assign b_write_data[i] = write_data[(((i + 1) *(DATA_WIDTH))-1) -: DATA_WIDTH];
end 

for (i=0; i < NUM_FIFO; i=i+1) begin : FIFOS
	fifo  #(DATA_WIDTH, Q_DEPTH_BITS, Q_IN_BUFFERS) U (
		clock,
		reset,
		b_write_data[i], 
		write[i], 
		b_read[i],
		1'b0, 

		b_read_data[i], 
		b_valid[i],			
		full[i], 
		empty[i] 
	);
end
endgenerate

reg [FIFO_BITS-1:0] turn, prev_turn; 
integer u; 

//--------------Code Starts Here----------------------- 
always @ (posedge clock) begin
	if (reset) begin 
		turn      <= 0; 
		prev_turn <= 0; 

		for(u= 0; u < NUM_FIFO; u = u+1) begin
			b_read[u] <= 0; 
		end 
	end 
	else begin 
		turn    <=  get_fifo(empty);
		if((~empty[turn]) & read) begin 
			if (prev_turn == turn) begin 
				b_read[turn]      <= 1; 
			end 
			else begin 
				b_read[turn]      <= 1; 
				b_read[prev_turn] <= 0;
				prev_turn         <= turn;
			end 
		end 
		else begin 
			b_read[prev_turn] <= 0;
		end 
	end             
end

//----------------------------------------------------
// Drive the outputs
//----------------------------------------------------
	assign read_data   = (b_valid[prev_turn])? b_read_data [prev_turn]: 0; 
	assign out_fifo    =  prev_turn;
    assign valid       = (b_valid[prev_turn])? 1 : 0;
	
endmodule
